@@ -33,7 +33,7 @@
 (cl:defpackage #:iterate.test
   (:use #:cl #:iterate
 	#+sbcl #:sb-rt
-	#-sbcl #:rt))
+	#-sbcl #:regression-test))
 
 (cl:in-package #:iterate.test)
 
@@ -51,6 +51,12 @@
       (list x y))
   (4 6))
 
+(deftest dsetq.3
+    (let ((a '((1 2) 3)) b)
+      (dsetq (a b) a)
+      (values a b))
+  (1 2) 3)
+
 (deftest dsetq.destructuring.1
     (let (x y)
       (dsetq (x . y) (list* 4 5 6))
@@ -81,16 +87,21 @@
       x)
   (b . c))
 
+(deftest dsetq.values.4
+    (let (x y z) (dsetq (values nil x (y z))
+			(values 1 'a '(b c))) (list x y z))
+  (a b c))
+
 (deftest repeat.1
     (iter (repeat 9) (count 1))
   9)
 
 (deftest repeat.2
-    (iter (repeat 2.5s0) (count t))
+    (iter (repeat 2.5s0) (counting t))
   3)
 
 (deftest repeat.3
-    (iter (repeat -1.5f0) (count t))
+    (iter (repeat -1.5f0) (counting t))
   0)
 
 (deftest locally.1
@@ -101,7 +112,7 @@
 (deftest locally.2
     (iterate (for i in '(1 2 3)) (repeat 2)
 	     (locally
-	       (declare (optimize (safety)))
+	       (declare (optimize safety))
 	       (declare (fixnum i)) (collect i)))
   (1 2))
 
@@ -142,6 +153,15 @@
     (iter (repeat 3) (thereis nil) (finally (return 2)))
   2)
 
+(deftest thereis.finally-protected.1
+    (iter (repeat 3) (thereis 4) (finally-protected (prin1 "hi")))
+  4)
+
+(deftest thereis.finally-protected.2
+    (with-output-to-string (*standard-output*) 
+      (iter (repeat 3) (thereis 4) (finally-protected (princ "hi"))))
+  "hi")
+
 (deftest finding.such-that.2
     (iter (for i in '(7 -4 2 -3 4))
 	  (if (plusp i)
@@ -153,6 +173,7 @@
     (iter (for i in '(1 2 3))
 	  (finding (1+ i)
 		   such-that #'(lambda (x)
+				 (declare (ignore x))
 				 (collect (- i) into m))))
   2) ; not -1 as some old version did
 
@@ -161,7 +182,7 @@
 	  (finding (1+ i)
 		   such-that #'(lambda (x)
 				 (finding (- x)
-					  such-that #'(lambda (x) nil)
+					  such-that #'(lambda (x) x nil)
 					  into n)
 				 t)
 		   into m)
@@ -205,7 +226,7 @@
   nil)
 
 (defun setup-hash-table (hash)
-  (dotimes (i (random 100))
+  (dotimes (i (random 100)) (declare (ignorable i))
 	(setf (gethash (random 10000) hash) (random 10000))
 	(setf (gethash (gensym) hash) (gensym))))
 
@@ -215,7 +236,7 @@
 	   (generated-entries
 	    (iterate (for (key) in-hashtable hash)
 		     (collect key))))
-      (maphash #'(lambda (key value) (push key all-entries)) hash)
+      (maphash (lambda (key value) value (push key all-entries)) hash)
       (= (length all-entries)
 	 (length generated-entries)
 	 (length (union all-entries generated-entries
@@ -226,9 +247,7 @@
     (let ((items nil)
 	  (hash (make-hash-table)))
       (setup-hash-table hash)
-      (maphash (lambda (key item)
-		 (push item items))
-	       hash)
+      (maphash (lambda (key item) key (push item items)) hash)
       (set-difference items
 		      (iterate (for (key item) in-hashtable hash)
 			       (declare (ignore key))
@@ -239,9 +258,7 @@
     (let ((items nil)
 	  (hash (make-hash-table)))
       (setup-hash-table hash)
-      (maphash (lambda (key item)
-		 (push item items))
-	       hash)
+      (maphash (lambda (key item) key (push item items)) hash)
       (set-difference items
 		      (iterate (for (nil item) in-hashtable hash)
 			       (collect item))))
@@ -350,6 +367,8 @@
   (1 10 ()))
 
 (deftest in-stream.2
+    ;; This fails in cmucl, sbcl and gcl, because string-input-streams
+    ;; are always open, even after close.
     (let ((s (make-string-input-stream "(")))
       (ignore-errors
        (iter (for x in-stream s :using #'read)))
@@ -362,6 +381,18 @@
 	  (accumulating (digit-char-p c) by #'+ initial-value 0))
   10)
 
+(deftest in-stream.reducing
+    (iter (with s = (make-string-input-stream "(+ 2)(+ 10)(- 5)(+ 6)"))
+	  (for (op num) in-stream s)
+	  (reducing num :by op :initial-value 0))
+  13)
+
+(deftest in-stream.accumulating
+    (iter (with s = (make-string-input-stream "(+ 2)(+ 10)(- 5)(+ 6)"))
+	  (for (op num) in-stream s)
+	  (accumulating num :by op :initial-value 0))
+  -1) ; (6 + (5 - (10 + (2 + 0))))
+
 (deftest in-stream.generate
     (iter (with s = (make-string-input-stream "2 + 10 - 5 + 6"))
       (with start = (read s))
@@ -463,6 +494,12 @@
 	     (maximize (length elt)))
   #.(reduce #'max *list-of-lists* :key #'length))
 
+(deftest finding.minimizing.1
+    (iterate (for elt in *list-of-lists*)
+	     (finding elt minimizing #'length into (e m))
+	     (finally (return m)))
+  #.(reduce #'min *list-of-lists* :key #'length))
+
 (deftest minimize.1
     (iterate (for elt in *list-of-lists*)
 	     (minimizing (length elt) into m)
@@ -540,6 +577,16 @@
 	  (accumulate 1 by #'cons initial-value 2))
   2)
 
+(deftest in-string.downto.1
+    (iter (for c in-string "235" downto 1)
+	  (accumulate (digit-char-p c) by 'cons))
+  (3 5))
+
+(deftest in-sequence.downto.1
+    (iter (for c in-sequence "235" downto 1)
+	  (accumulate (digit-char-p c) by #'cons))
+  (3 5))
+
 (deftest reducing.1
     (iter (for c in-string "235")
 	  (reducing (digit-char-p c) by 'list initial-value 1))
@@ -560,6 +607,16 @@
 	  (reducing i by #'- :initial-value '0))
   -12)
 
+(deftest reducing.5
+    (iter (for x in-vector #(3))
+	  (reducing (cons x x) by #'list))
+  (3 . 3))
+
+(deftest reducing.6
+    (iter (for x in-vector (vector 3))
+	  (reducing (cons x x) by #'list :initial-value nil))
+  (nil (3 . 3)))
+
 
 ;; synonyms (e.g. GENERATING, COLLECTING) didn't work
 
@@ -567,19 +624,19 @@
     (iter (generate (key . item) in '((a . 1) (b . 2) (c .3)))
 	  (collect (next key))
 	  (collect (next item)))
-  (A 2 C))
+  (a 2 c))
 
 (deftest generating.destructuring.1
     (iter (generating (key . item) in '((a . 1) (b . 2) (c .3)))
 	  (collect (next key))
 	  (collect (next item)))
-  (A 2 C))
+  (a 2 c))
 
 (deftest for.generate-t.destructuring.1
     (iter (for (key . item) in '((a . 1) (b . 2) (c .3)) :generate t)
 	  (collect (next key))
 	  (collect (next item)))
-  (A 2 C))
+  (a 2 c))
 
 (deftest generate.next.1
     (iter (generate c in '(a b c d e f g h i j k l m n o p q))
@@ -654,6 +711,34 @@
 	  (adjoining (next i)))
   (a 0 b 0 c 0))
 
+(deftest after-each.2
+    (iter (with i = 0)
+	  (while (< i 4))
+	  (after-each (incf i)) ; the C programmer's for (;;) loop
+	  (collect i))
+  (0 1 2 3))
+
+(deftest after-each.3
+    (iter (with i = 0)
+	  (while (< i 4))
+	  (collect i)
+	  (after-each (incf i)))
+  (0 1 2 3))
+
+(deftest next-iteration.1
+    (iter (for i below 10)
+	  (when (oddp i) (next-iteration))
+	  (count t))
+  5)
+
+(deftest next-iteration.2
+    (iter (for thing in '(var &optional else &key (test #'eql)))
+	  (collect
+	   (cond ((consp thing) (first thing))
+		 ((not (member thing lambda-list-keywords)) thing)
+		 (t (next-iteration)))))
+  (var else test))
+
 ;;;; tests from the documentation:
 
 (deftest collect.2
@@ -680,7 +765,7 @@
 	    (print "I will not talk in class.")))
   #.(with-output-to-string (*standard-output*)
       (dotimes (i 100)
-	(declare (ignore i))
+	(declare (ignorable i)) ; cmucl/sbcl complain about (ignore i)
 	(print "I will not talk in class."))))
 
 ;;; for.next.1 and for.do-next.1 used to be broken in older versions;
@@ -691,7 +776,7 @@
 	     (finally (return i)))
   11)
 
-;;; This gave STYLE-WARNINGS for undefined I in old versions.
+;;; This gave STYLE-WARNINGS for undefined i in old versions.
 (deftest for.do-next.1
     (iterate (initially (setq i 0))
 	     (as i do-next (if (> i 10) (terminate) (incf i)))
@@ -699,12 +784,46 @@
   11)
 
 (deftest for.do-next.2
-    ;; INITIALLY not needed because it is inferred from type declaration
+    ;; INITIALLY not needed because 0 is inferred from type declaration
     (iterate (for i do-next (if (> i 7) (terminate) (incf i)))
 	     (declare (type fixnum i))
 	     (finally (return i)))
   8)
 
+(deftest for.do-next.3
+    (iter (for a from 1 to 3)
+	  (for b = (1+ (* a a)))
+	  ;; (values ...) is supported, even though (x y) would do
+	  (for (values x y) do-next (dsetq (values x y) (floor b a)))
+	  (collect x) (collect y))
+  (2 0 2 1 3 1))
+
+(deftest for.next.walk
+    (iter (repeat 2)
+	  (for x next (progn (after-each (collect 1)) 2))
+	  (collect x))
+  (2 1 2 1))
+
+(deftest for.do-next.walk
+    (iter (repeat 2)
+	  (for x do-next (progn (after-each (collect 1)) (dsetq x 2)))
+	  (collect x))
+  (2 1 2 1))
+
+(deftest for.next.previous
+    (iter (for i from 2 to 4)
+	  (for x next (progn (after-each (collect i)) (- i)))
+	  (for z previous x initially 0)
+	  (nconcing (list z x)))
+  (0 -2 2 -2 -3 3 -3 -4 4))
+
+(deftest for.do-next.previous
+    (iter (for i from 2 to 4)
+	  (for x do-next (progn (setq x (- i)) (after-each (collect i))))
+	  (for z previous x initially 0)
+	  (appending (list z x)))
+  (0 -2 2 -2 -3 3 -3 -4 4))
+
 (deftest for-nongenerator.1
     (iter (for el in '(a b c d))
 	  (generate i upfrom 1)
@@ -714,7 +833,7 @@
 	  (if el (collect (cons el i)))))
   
 
-(deftest for.previous.1
+(deftest for.previous.in
     (iter (for el in '(1 2 3 4))
 	  (for pp-el previous el back 2 initially 0)
 	  (collect pp-el))
@@ -757,6 +876,13 @@
 	  (collect i))
   (20 21 22 23 24))
 
+(deftest for.initially.1
+    (iter (with (v z))
+	  (for i initially (length v) then (1+ i))
+	  (collect (cons i z))
+	  (while (evenp i)))
+  ((0) (1)))
+
 (deftest sum.1
     (iter (for el in '(100 200 300))
 	  (sum el into x)
@@ -858,7 +984,7 @@
 	  collect (sqrt i)))
 
 (deftest collect.type.string.1
-    (locally (declare safety (debug 2) (speed 0) (space 1))
+    (locally (declare (optimize safety (debug 2) (speed 0) (space 1)))
       (iter (declare (iterate:declare-variables))
 	    (for s in-vector '#(\a |b| |cD|))
 	    (collect (char (symbol-name s) 0) :result-type string)))
@@ -870,13 +996,55 @@
 	    (collect c :result-type string)))
   "aBc")
 
-(deftest collect.type.vector
-    (locally (declare safety (debug 2) (speed 0) (space 1))
+(deftest collect.type.string.3
+    (iter (for c in-string "235" downfrom 1)
+	  (collect c into s result-type string)
+	  (finally (return s)))
+  "32")
+
+(deftest collect.type.vector.1
+    (locally (declare (optimize safety (debug 2) (speed 0) (space 1)))
       (iter (declare (iterate:declare-variables))
 	    (for s in-vector '#(\a |b| |cD|))
 	    (collect (char (symbol-name s) 0) :result-type vector)))
   #(#\a #\b #\c))
 
+(deftest collect.type.vector.2
+    (iter (for c in-vector "235" downfrom 1)
+	  (collect (digit-char-p c) :into v :result-type vector)
+	  (finally (return v)))
+  #(3 2))
+
+(deftest adjoin.1
+    (iter (for i in '("ab" "aB" "abc" "AB"))
+	  (adjoining i at :start :test #'string-equal))
+  ("abc" "ab"))
+
+(deftest adjoin.2
+    (iter (for i in '("ab" "aB" "abc" "AB"))
+	  (adjoining i at :start))
+  ("AB" "abc" "aB" "ab"))
+
+(deftest adjoin.3
+    (iter (for i in '("ab" "aB" "abc" "AB"))
+	  (adjoining i :at end #:test #'string-equal))
+  ("ab" "abc"))
+
+(deftest adjoin.4
+    (iter (for i in '("ab" "aB" "abc" "AB"))
+	  (adjoining i at :end))
+  ("ab" "aB" "abc" "AB"))
+
+(deftest adjoin.5
+    (iter (for i in '("ab" "aB" "abc" "AB"))
+	  (adjoining (string-downcase i) at :start :test #'string-equal))
+  ("abc" "ab"))
+
+(deftest adjoin.6
+    (iter (for i in '("ab" "aB" "abc" "AB"))
+	  (adjoining (string-upcase i) #:at :end test #'string=))
+  ("AB" "ABC"))
+
 (deftest append.1
     (iter (for l on '(1 2 3))
 	  (appending l at :start))
@@ -1013,6 +1181,16 @@
 	  (finally (return x)))
   -8)
 
+(deftest walk.multiple-value-bind
+    (string-upcase
+     (iter (for name in-vector (vector 'iterate "FoOBaRzOt" '#:repeat))
+	   (multiple-value-bind (sym access)
+	       (find-symbol (string name) #.*package*)
+	     (declare (type symbol sym))
+	     (collect (if access (char (symbol-name sym) 0) #\-)
+		      result-type string))))
+  "I-R")
+
 (deftest subblocks.1
     (iter fred
 	  (for i from 1 to 10)
@@ -1097,7 +1275,7 @@
 (deftest type.7
     (iter (declare (iterate:declare-variables))
 	  (repeat 0)
-	  (minimize (the double-float '1)))
+	  (minimize (the double-float '1.0d0)))
   0.0d0)
 
 (deftest static.error.1
@@ -1274,34 +1452,32 @@
       (iter::remove-clause '(multiply.clause &optional INTO))))
   nil)
 
-(iter:defmacro-clause (FOR var IN-WHOLE-VECTOR.clause v)
+(iter:defmacro-clause (for var IN-WHOLE-VECTOR.clause v)
   "All the elements of a vector (disregards fill-pointer)"
-  (let ((vect (gensym))
-        (end (gensym))
-        (index (gensym)))
+  (let ((vect (gensym "VECTOR"))
+        (index (gensym "INDEX")))
     `(progn
        (with ,vect = ,v)
-       (with ,end = (array-dimension ,vect 0))
-       (for ,index from 0 below ,end)
-       (dsetq ,var (aref ,vect ,index)))))
+       (for ,index from 0 below (array-dimension ,vect 0))
+       (for ,var = (aref ,vect ,index)))))
 
 (deftest in-whole-vector.clause
-    (iter (for i in-whole-vector.clause (make-array '(3) :fill-pointer 2
-					     :initial-contents '(1 2 3)))
+    (iter (for i IN-WHOLE-VECTOR.clause (make-array 3 :fill-pointer 2
+					 :initial-contents '(1 2 3)))
 	  (collect i))
   (1 2 3))
 
 (deftest in-vector.fill-pointer
-    (iter (for i in-vector (make-array '(3) :fill-pointer 2
-				       :initial-contents '(1 2 3)))
+    (iter (for i in-vector (make-array 3 :fill-pointer 2
+			    :initial-contents '(1 2 3)))
 	  (collect i))
   (1 2))
 
-(iter:defmacro-driver (FOR var IN-WHOLE-VECTOR v)
+(iter:defmacro-driver (for var IN-WHOLE-VECTOR v)
   "All the elements of a vector (disregards fill-pointer)"
-   (let ((vect (gensym))
-         (end (gensym))
-         (index (gensym))
+   (let ((vect (gensym "VECTOR"))
+         (end (gensym "END"))
+         (index (gensym "INDEX"))
          (kwd (if iter:generate 'generate 'for)))
      `(progn
         (with ,vect = ,v)
@@ -1312,13 +1488,13 @@
                                (aref ,vect ,index))))))
 
 (deftest in-whole-vector.driver
-    (iter (for i in-whole-vector (make-array '(3) :fill-pointer 2
+    (iter (for i IN-WHOLE-VECTOR (make-array '(3) :fill-pointer 2
 					     :initial-contents '(1 2 3)))
 	  (collect i))
   (1 2 3))
 
 (deftest in-whole-vector.generate
-    (iter (generating i in-whole-vector (make-array '(3) :fill-pointer 2
+    (iter (generating i IN-WHOLE-VECTOR (make-array '(3) :fill-pointer 2
 						    :initial-contents '(1 2 3)))
 	  (collect (next i)))
   (1 2 3))
@@ -1327,7 +1503,7 @@
     (progn
       (iter:defclause-sequence IN-WHOLE-VECTOR.seq INDEX-OF-WHOLE-VECTOR
 	:access-fn 'aref
-	:size-fn #'(lambda (v) (array-dimension v 0))
+	:size-fn '#'(lambda (v) (array-dimension v 0))
 	:sequence-type 'vector
 	:element-type t
 	:element-doc-string 
@@ -1338,16 +1514,43 @@
   t)
 
 (deftest in-whole-vector.seq
-    (iter (for i in-whole-vector.seq (make-array '(3) :fill-pointer 2
+    (iter (for i IN-WHOLE-VECTOR.seq (make-array '(3) :fill-pointer 2
 						 :initial-contents '(1 2 3)))
 	  (collect i))
   (1 2 3))
 
+(deftest in-whole-vector.seq.index
+    (iter (for i INDEX-OF-WHOLE-VECTOR
+	       (make-array 3 :fill-pointer 2 :initial-contents '(1 2 3)))
+	  (for j previous i :initially 9)
+	  (collect (list j i)))
+  ((9 0)(0 1)(1 2)))
+
+(deftest in-whole-vector.seq.with-index
+    (iter (for e IN-WHOLE-VECTOR.seq
+	       (make-array '(3) :fill-pointer 2 :initial-contents '(a b c))
+	       :with-index i)
+	  (for j previous i :initially 9)
+	  (collect (list j i e)))
+  ((9 0 a)(0 1 b)(1 2 c)))
+
+(deftest in-whole-vector.seq.generate
+    (iter (generate e IN-WHOLE-VECTOR.seq
+	       (make-array 3 :fill-pointer 2 :initial-contents '(a b c))
+	       :with-index i)
+	  (collect (list (next e) e i)))
+  ((a a 0) (b b 1) (c c 2)))
+
 (deftest display.1
     (let ((*standard-output* (make-broadcast-stream)))
       (display-iterate-clauses) t)
   t)
 
+(deftest display.2
+    (let ((*standard-output* (make-broadcast-stream)))
+      (display-iterate-clauses 'for) t)
+  t)
+
 (deftest multiple-value-prog1.1
     (iter (for x in '(a b c))
 	  (collect (multiple-value-prog1 7)))
@@ -1365,7 +1568,7 @@
 
 (deftest handler-bind.1
     (iter (for i from -1 to 2 by 2)
-	  (handler-bind ((error (lambda(c) nil)))
+	  (handler-bind ((error (lambda(c) c nil)))
 	    (collect i)))
   (-1 1))
 
@@ -1422,11 +1625,15 @@
   (1 2 3))
 
 (deftest bug/macrolet.2
-    (prog1
-     (ignore-errors ; would yield 1 if correct
-      (iterate (repeat 10)
-	       (macrolet ((foo () 1))
-		 (multiplying (foo))))))
+    (progn
+      (format *error-output*
+	      "~&Note: These tests generate warnings ~
+  involving MACROLET within Iterate~%")
+      (values
+       (ignore-errors ; would yield 1 if correct
+	 (iterate (repeat 10)
+		  (macrolet ((foo () 1))
+		    (multiplying (foo)))))))
   nil)
 
 (deftest macrolet.3
@@ -1435,32 +1642,33 @@
 			    (foo))))
   1)
 
-;; related to the above. hashtable iterators are specified to be
-;; defined as macrolets
-(deftest bug/hashtable-walk.1
+;; Hashtable iterators are specified to be defined as macrolets.
+;; But we handle these by special-casing with-hash-table/package-iterator
+(deftest nested-hashtable.1
     (let ((ht1 (make-hash-table))
           (ht2 (make-hash-table)))
       (setup-hash-table ht2)
       (setf (gethash 'a ht1) ht2)
-      (values
-       (ignore-errors
-	(iter outer (for (k1 v1) in-hashtable ht1)
-	      (iter (for (k2 v2) in-hashtable ht2)
-		    (in outer (collect k2)))))))
-  nil)
+      (= (hash-table-count ht2)
+	 (length
+	  (iter outer (for (k1 v1) in-hashtable ht1)
+		(iter (for (k2 v2) in-hashtable ht2)
+		      (in outer (collect k2)))))))
+  t)
 
-(deftest macrolet.in-hashtable.1
+(deftest nested.in-hashtable.2
+    ;; Here the inner macrolet code does not affect the outer iteration
     (let ((ht1 (make-hash-table))
           (ht2 (make-hash-table)))
       (setup-hash-table ht2)
       (setf (gethash 'a ht1) ht2)
       (iter (for (k1 v1) in-hashtable ht1)
-	    (counting ; the inner macro is not walked in this case
+	    (counting
 	     (iter (for (k2 nil) in-hashtable ht2)
 		   (count k2)))))
   1)
 
-(deftest macrolet.in-hashtable.2
+(deftest nested.in-hashtable.3
     (let ((ht1 (make-hash-table))
           (ht2 (make-hash-table)))
       (setup-hash-table ht2)
@@ -1472,7 +1680,17 @@
 	     (collect k1))))
   (a))
 
-(deftest macrolet.loop
+(deftest nested.in-package
+  (< 6
+     (print
+      (iter (for scl in-package '#:iterate :external-only t)
+	    (count ; Iterate exports ~50 symbols
+	     (iter (for si in-package #.*package*)
+		   (thereis (eq si scl))))))
+     80)
+  t)
+
+(deftest macrolet.loop-finish
     (iter (for l on *an-alist*)
 	  (loop for e in l
 		when (equal (car e) 'zero)
